from dataclasses import dataclass
from typing import Any, Dict, List

import requests
from requests import Session

from ..asset_types.edstem import Announcement
from . import announ
from .canvas import tab


@dataclass
class Checker(announ.Checker, tab.Checker):
    api_url: str = None
    api_headers: Dict = None

    @property
    def pin_amount_path(self):
        return self.content_path.with_stem(self.content_path.stem + "pinned")

    def get_api_data(self, limit=1, sort="new", offset=0, filter="staff", **params):
        params |= {"limit": limit, "sort": sort, "offset": offset, "filter": filter}
        return requests.get(
            self.api_url, params=params, headers=self.api_headers
        ).json()["threads"]

    def get_thread(self, position):
        threads = self.get_api_data(offset=position)
        thread = threads[0] if threads else None
        return thread

    def get_items(self) -> List[Any]:
        self.prepare_api()
        items = [*self.get_thread_items(), *self.get_unpinned_items()]
        return items

    def get_thread_items(self, start_offset=0):
        last_item_new = True
        offset = start_offset
        while last_item_new:
            thread = self.get_thread(offset)
            if thread and thread["id"] not in self.old_content:
                yield thread
                offset += 1
            else:
                last_item_new = False

    def get_unpinned_items(self):
        unpinned_index = int(self.pin_amount_path.text or "0")
        thread = self.get_thread(unpinned_index)

        while thread and thread["is_pinned"]:
            unpinned_index += 1
            thread = self.get_thread(unpinned_index)

        if thread and thread["id"] not in self.old_content:
            yield thread

        self.pin_amount_path.text = unpinned_index
        yield from self.get_thread_items(start_offset=unpinned_index)

    def make_item(self, item):
        item_info = item | dict(save_folder=self.save_folder)
        return Announcement.from_dict(item_info)

    @classmethod
    def tab_name(cls):
        return "Ed Discussion"

    def prepare_api(self):
        url = "https://us.edstem.org/api/login_token"
        course_id, login_token = self.get_tokens()
        self.api_url = f"https://us.edstem.org/api/courses/{course_id}/threads"

        data = {"login_token": login_token}
        resp = requests.post(url, json=data).json()
        self.api_headers = {"x-token": resp["token"]}

    def get_tokens(self):
        session = Session()
        launch_url = self.get_launch_url()
        launch_response = session.get(launch_url).text
        relocation_url = None

        for _ in range(2):
            url, data = self.parse_form(launch_response)
            response = session.post(url, data=data)
            relocation_url = response.url
            launch_response = response.text

        info = relocation_url.split("/")[-1]
        tokens = info.split("?_logintoken=")
        return tokens



from libs.threading import Threads

from canvasdl.checkers import get_checkers
from canvasdl.ui.userinterface import UserInterface

from . import courseinfo


def check_changes():
    courses = courseinfo.get_courses()
    checkers = get_checkers()
    course_checkers = [
        checker.Checker(course) for course in courses for checker in checkers
    ]
    course_checkers = [checker for checker in course_checkers if checker.should_check()]
    content_threads = Threads([c.check for c in course_checkers]).start()
    UserInterface.show_progress(len(course_checkers))
    content_threads.join()



from __future__ import annotations

from dataclasses import dataclass
from typing import List

import cli
import downloader
import m3u8

from ..utils import Path
from . import base


@dataclass
class RelativeSegment(base.Item):
    RelativeStart: float


@dataclass
class Stream(base.Item):
    RelativeStart: float
    RelativeEnd: float
    StreamUrl: str
    RelativeSegments: List[RelativeSegment] | None

    @property
    def duration(self):
        return self.RelativeEnd - self.RelativeStart

    @property
    def to_download(self):
        return self.duration > 20

    @property
    def start(self):
        return (
            self.RelativeSegments[0].RelativeStart
            if self.RelativeSegments
            else self.RelativeStart
        )


@dataclass
class DownloadStream(Stream):
    index: int
    streams: Streams

    @property
    def name(self):
        name = self.streams.name
        if self.index > 0:
            name += f"__VIEW__{self.index + 1}"
        return name

    @property
    def dest(self):
        return (self.streams.folder / self.name).with_suffix(".mp4")

    def download(self):
        if self.to_download and not self.dest.exists():
            self.start_download()

    def start_download(self):
        if "m3u8" in self.StreamUrl:
            download_m3u8(self.StreamUrl, self.dest)
        else:
            downloader.download(self.StreamUrl, self.dest)
        self.synchronize_download_start()
        self.dest.mtime = self.streams.mtime

    def synchronize_download_start(self):
        offset_difference = self.streams.start - self.start
        cut(self.dest, offset_difference)


@dataclass
class Streams(base.Item):
    streams: List[Stream]
    folder: Path
    name: str
    mtime: float

    @property
    def streams_to_download(self):
        for stream in self.streams:
            if stream.to_download:
                yield stream

    @property
    def start(self):
        return max([s.start for s in self.streams_to_download])

    def download(self):
        for i, stream in enumerate(self.streams_to_download):
            stream_info = stream.dict() | dict(streams=self, index=i)
            stream = DownloadStream.from_dict(stream_info)
            stream.download()
        if self.folder.exists():
            self.folder.mtime = self.mtime


def download_m3u8(url: str, dest: Path, headers=None):
    headers = headers or {}
    playlist = m3u8.load(url, headers=headers)
    playlist = m3u8.load(playlist.playlists[0].absolute_uri, headers=headers)
    url = playlist.segments[0].absolute_uri
    downloader.download(url, dest)


def cut(dest: Path, start_offset):
    with Path.tempfile().with_suffix(dest.suffix) as tmp:
        args = (
            "ffmpeg",
            "-i",
            dest,
            "-ss",
            start_offset,
            "-vcodec",
            "copy",
            "-acodec",
            "copy",
            tmp,
        )
        cli.get(args)
        tmp.copy_to(dest)



from datetime import datetime, time, timedelta

from gcsa.event import Event
from gcsa.google_calendar import GoogleCalendar

from . import configchecker
from .config import config
from .path import Path

DAY = timedelta(days=1)


def add_todo(message, date):
    use_calendar = configchecker.google_calendar_credentials_valid()
    is_future_event = date.replace(tzinfo=None) > datetime.now()
    if use_calendar and is_future_event:
        add_event(message, date, date)


def add_event(message, start, end):
    start_time = start.time()
    full_day = start_time in (time(hour=23, minute=59), time(hour=0, minute=0))
    if full_day:
        # start quarter before midnight
        subtract_minutes = 15 if start_time.minute == 0 else 14
        start -= timedelta(hours=0, minutes=subtract_minutes)

    args = (config.google_calendar_id,) if config.google_calendar_id else ()
    calendar = GoogleCalendar(*args, credentials_path=Path.calendar_credentials)
    existing_same_events = list(
        calendar.get_events(time_min=start, time_max=start + DAY, query=message)
    )
    if not existing_same_events:
        event = Event(message, start=start, end=end, colorId=11)
        calendar.add_event(event)



from . import canvas, drive, edstem, gradescope, piazza, website


def get_checkers():
    return (
        *canvas.get_checkers(),
        drive,
        edstem,
        gradescope,
        piazza,
        website,
    )



import html
from abc import ABC
from dataclasses import dataclass
from datetime import datetime
from typing import Any

import requests
from bs4 import BeautifulSoup

from ... import utils
from ...utils import time
from . import base


@dataclass
class Due:
    name: str
    date: datetime

    @property
    def title(self):
        return f"{self.name} ({time.export_datetime(self.date)})"


@dataclass
class Checker(base.Checker, ABC):
    tab: Any = None

    @classmethod
    def tab_name(cls):
        raise NotImplementedError

    def should_check(self):
        tabs = [tab for tab in self.api.get_tabs() if tab.label == self.tab_name()]
        self.tab = tabs[0] if tabs else None
        return self.tab is not None

    def get_launch_url(self):
        launch_request_url = self.tab.url
        params = {"access_token": utils.config.API_KEY}
        launch_url = requests.get(launch_request_url, params=params).json()["url"]
        return launch_url

    def get_redirected_content(self):
        launch_url = self.get_launch_url()
        launch_response = requests.get(launch_url).text
        url, data = self.parse_form(launch_response)
        return requests.post(url, data=data).text

    @classmethod
    def parse_form(cls, html_string: str):
        soup = BeautifulSoup(html_string, features="lxml")
        form = soup.find("form")
        url = form["action"]
        process = html.unescape
        data = {
            process(field["name"]): process(field["value"])
            for field in form.find_all("input")
        }
        return url, data



from dataclasses import dataclass
from datetime import datetime
from typing import Any

import dateutil.parser

from ..asset_types import SaveItem
from ..asset_types.canvas import Asset
from ..utils import calendar, time
from .course import Course


@dataclass
class Assignment(Asset, SaveItem):
    created_at: str
    created_at_date: datetime
    updated_at: str
    updated_at_date: datetime
    lock_at: str | None
    unlock_at: str | None
    description: str | None
    due_at: str | None
    points_possible: float
    grading_type: Any
    assignment_group_id: int
    grading_standard_id: int | None
    peer_reviews: Any
    automatic_peer_reviews: Any
    position: int
    grade_group_students_individually: Any
    anonymous_peer_reviews: Any
    group_category_id: Any
    post_to_sis: Any
    moderated_grading: Any
    omit_from_final_grade: Any
    intra_group_peer_reviews: Any
    anonymous_instructor_annotations: Any
    anonymous_grading: Any
    graders_anonymous_to_graders: Any
    grader_count: Any
    grader_comments_visible_to_graders: Any
    final_grader_id: Any
    grader_names_visible_to_final_grader: Any
    allowed_attempts: Any
    annotatable_attachment_id: Any
    secure_params: Any
    lti_context_id: Any
    course_id: Any
    name: Any
    submission_types: Any
    has_submitted_submissions: Any
    due_date_required: Any
    max_name_length: Any
    in_closed_grading_period: Any
    graded_submissions_exist: Any
    is_quiz_assignment: Any
    can_duplicate: Any
    original_course_id: Any
    original_assignment_id: Any
    original_lti_resource_link_id: Any
    original_assignment_name: Any
    original_quiz_id: Any
    workflow_state: Any
    important_dates: Any
    muted: bool
    html_url: str
    published: bool
    only_visible_to_overrides: Any
    locked_for_user: bool
    submissions_download_url: Any
    post_manually: Any
    anonymize_students: Any
    require_lockdown_browser: Any
    due_at_date: datetime = None
    unlock_at_date: datetime = None
    lock_at_date: datetime = None
    allowed_extensions: Any = None
    quiz_id: int = None
    anonymous_submissions: bool = False
    lock_explanation: Any = None
    lock_info: Any = None

    course: Course = None

    def save(self):
        if self.due_at:
            due = dateutil.parser.parse(self.due_at)
            message = self.course.assignment_name(self.name)
            calendar.add_todo(message, date=due)

    @property
    def display_title(self):
        time_str = time.export_time(self.due_at) if self.due_at else ""
        return f"{self.name} ({time_str})"

    @property
    def save_id(self):
        return self.id



import time

from . import starter


def main():
    hours = 2
    seconds = hours * 60**2
    while True:
        time.sleep(seconds)
        starter.check_changes()


if __name__ == "__main__":
    main()



from dataclasses import dataclass, field

from canvasdl.utils.path import Path

from ..utils import config
from .base import Item


@dataclass(order=True)
class Course(Item):
    name: str
    id: str
    video_id: str = None
    drive_name: str = None
    piazza_id: str = None
    websites: dict = field(default_factory=dict)

    @property
    def sort_index(self):
        return -Path.content_path(self.name).size

    @property
    def abbreviation(self):
        removes = ("Topics in", "Introduction to", "and")
        name = self.name
        for r in removes:
            name = name.replace(r, "")
        return "".join(word[0].capitalize() for word in name.split(" ") if word)

    def assignment_name(self, name: str):
        return f"{self.abbreviation}: {name} due"


@dataclass
class Courses:
    courses: list[Course]

    @classmethod
    def from_dict(cls, courses):
        if config.one_course_nr:
            courses = [courses[config.one_course_nr - 1]]
        return Courses(sorted([Course(**c) for c in courses]))



from functools import cached_property

import requests
import video_exporter

from canvasdl.utils import Path

from ...asset_types import SavedVideo, Video
from . import tab


class Checker(tab.Checker):
    @classmethod
    def tab_name(cls):
        return "Video Library"

    @property
    def api_url(self):
        return "https://cvn.hosted.panopto.com/Panopto/api/v1/"

    @property
    def path(self):
        return super().path / "Videos.html"

    def __post_init__(self):
        super().__post_init__()
        self.save_folder = Path.content_path(self.course.name, ("videos",)).with_suffix(
            ""
        )

    def make_item(self, item):
        item |= dict(save_folder=self.save_folder)
        return Video.from_dict(item)

    def get_items(self):
        session = self.authenticated_session
        folder_id = self.course.video_id

        if folder_id:
            url = f"{self.api_url}folders/{folder_id}/sessions"
            params = {"sortfield": "CreatedDate", "sortOrder": "Desc"}
            # pageNumber, https://demo.hosted.panopto.com/Panopto/api/docs
            # /index.html#/Folders/Folders_GetSessions
            response = session.get(url, params=params)
            content = response.json()["Results"]
        else:
            content = []
        return content

    def get_folder_id(self, name):
        folder_id = None
        if self.should_check():
            url = f"{self.api_url}folders/search"
            params = {"searchQuery": name}
            response = self.authenticated_session.get(url, params=params)
            results = response.json()["Results"]
            if results:
                folder_id = results[0]["Id"]

        return folder_id

    @cached_property
    def authenticated_session(self):
        launch_url = self.get_launch_url()
        launch_response = requests.get(launch_url).text
        url, data = self.parse_form(launch_response)
        session = requests.Session()
        session.post(url, data=data)

        uni = data["lis_person_sourcedid"]
        cookie = {
            f"CVNCanvas\\{uni}": (
                '{"lastSessionView":1,"sortData-lastSessionWithSearchSort":'
                '"{"column":3,"currentAscending":false}",'
                '"sortData-lastSessionSort":"{"column":1,"currentAscending":false}"}]'
            )
        }
        session.cookies.update(cookie)

        return session

    def export_downloads(self):
        self.download_recordings()
        self.export_html_files()

    def export_html_files(self):
        exports = {
            self.path: self.download_folder,
            Path.school / "Videos.html": list(self.download_folder.parent.iterdir()),
        }
        for dest, video_folders in exports.items():
            video_exporter.export(video_folders, dest, merge_folders=True)

    @property
    def download_folder(self):
        return (
            self.save_folder.parent.parent / "video_downloads" / self.save_folder.name
        )

    def download_recordings(self):
        for path in self.save_folder.iterdir():
            video = SavedVideo.from_dict(path.yaml)
            self.download_streams(video)

    def get_api(self, data):
        post_url = (
            "https://cvn.hosted.panopto.com/Panopto/Pages/Viewer/DeliveryInfo.aspx"
        )
        return self.authenticated_session.post(post_url, data=data).json()

    def download_streams(self, video: SavedVideo):
        info = self.get_api({"deliveryId": video.id, "responseType": "json"})
        video.download(info, self.download_folder)



from datetime import datetime

from dateutil import parser


def parse_time(time_string: str):
    return parser.parse(time_string).timestamp()


def export_time(time_str: str):
    return datetime.fromtimestamp(parse_time(time_str))


def export_datetime(date: datetime):
    return date.strftime("%d-%m-%Y %H:%M")


def export_mtime(mtime: float):
    date = datetime.fromtimestamp(mtime)
    return export_datetime(date)



from ...asset_types.assignment import Assignment
from . import base


class Checker(base.Checker):
    @property
    def path(self):
        return super().path / "Assignments"

    def get_items(self):
        return self.api.get_assignments()

    def make_item(self, item):
        assignment = Assignment.from_response(item)
        assignment.course = self.course
        return assignment



import threading
import time

from .progressbar import ProgressBar
from .silentpopup import SilentUIHandle


class Progress(ProgressBar):
    def __init__(self, message="Checking for new items", silent=False):
        try:
            from .popup import UIHandle  # noqa: autoimport
        except ModuleNotFoundError:
            silent = True
        super().__init__(title="School", message=message, silent=silent)
        self.messages = [message]
        self.auto_max = 0.0
        self.auto_add_value = 0.001

        self.do_auto_add = True
        self.auto_progress = 0

    def __enter__(self):
        super().__enter__()
        self.auto_add()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.do_auto_add = False
        super().__exit__(exc_type, exc_val, exc_tb)

    def add_message(self, message):
        self.messages.append(self.message)
        self.message = message

    def pop_message(self):
        if len(self.messages) > 0:
            message = self.messages.pop()
            self.message = message

    def auto_add(self):
        threading.Thread(target=self._auto_add).start()

    def _auto_add(self):
        while self.do_auto_add:
            if self.auto_progress < self.auto_max:
                self.auto_progress += self.auto_add_value
            time.sleep(0.005)

    def add_progress(self, value):
        self.progress += value

    @property
    def percentage(self):
        return max(super().percentage, self.auto_progress * 100)

    def reset(self):
        self.progress = 0


class ProgressManager:
    progress = Progress(silent=True)

    @classmethod
    def __enter__(cls):
        ProgressManager.progress = Progress()
        ProgressManager.progress.__enter__()

    @classmethod
    def __exit__(cls, exc_type, exc_val, exc_tb):
        ProgressManager.progress.__exit__(exc_type, exc_val, exc_tb)

    @classmethod
    @property
    def interactive_mode(cls):
        return not isinstance(cls.progress.handle, SilentUIHandle)



try:
    from .popup import Popup as ProgressBar

except ImportError:
    # no progress
    class ProgressBar:
        def __init__(self, *args, **kwargs):
            pass

        def __enter__(self):
            pass

        def __exit__(self, *_):
            pass

        def __getattr__(self, name):
            return self.empty

        def empty(self, *args, **kwargs):
            pass



from dataclasses import dataclass

import cli

from . import base


@dataclass
class Checker(base.Checker):
    def get_items(self):
        return []

    def should_check(self) -> bool:
        return bool(self.course.drive_name)

    def check_new_content(self):
        local_path = self.path / "Drive"
        remote_path = f"{self.course.drive_name}:"
        options = {"drive-export-formats": "pdf"}
        cli.run("rclone", "sync", options, remote_path, local_path)



import os
from dataclasses import dataclass
from functools import cached_property

import cli
from piazza_api import Piazza

from ..asset_types.piazza import Announcement
from . import announ


@dataclass
class Checker(announ.Checker):
    def should_check(self) -> bool:
        return bool(self.course.piazza_id)

    @cached_property
    def api(self):
        api = get_api()
        network_id = self.course.piazza_id
        api = api.network(network_id)
        return api

    def get_items(self):
        feed = self.api.get_feed(limit=999999, offset=0)
        ids = [post["id"] for post in feed["feed"]]
        new_ids = [i for i in ids if i not in self.old_content]
        return new_ids

    def make_item(self, item):
        item_info = self.api.get_post(item) | dict(save_folder=self.save_folder)
        return Announcement.from_dict(item_info)


def get_api():
    piazza = Piazza()
    email = os.environ["school_email"]
    password = cli.get("pw piazza")

    piazza.user_login(email=email, password=password)
    return piazza
